#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// helper for 'get_s_record'
static unsigned char get_char(char *ptr)
{
	unsigned char i, ret = 0;
	for (i = 0; i < 2; i++, *ptr++)
	{
		ret <<= 4;
		if ((*ptr & 0xf0) == 0x30 && (*ptr & 0xf) < 10) ret |= (*ptr & 0xf);
		if ((*ptr & 0xd0) == 0x40 && (*ptr & 0xf) > 0 && (*ptr & 0xf) < 7) ret |= ((*ptr & 0x7) + 9);
	}

	return ret;
}

// function to import easy68k 's' records into raw data
static int get_s_record(unsigned char *dest)
{
	char line[128];

	FILE *fa = fopen("puzlstar_asm.s68", "rt");
	if (fa == NULL) {
		printf ("Failed to open puzlstar_asm.s68.\n");
		return 1;
	}

	printf ("Importing \'S\' record...\n");

	while (1)
	{
		if (fgets (line, 128, fa) == NULL) break;

		char *ptr = line;
		while (*ptr == ' ' || *ptr == '\t') ptr++;
		if (*ptr != 'S') continue; // ??
		ptr++; // past 's' (1 byte)

		if (*ptr != '1' && *ptr != '2' && *ptr != '3') continue;

		unsigned char address_size = (*ptr & 3) + 1; ptr++;
		unsigned int address = 0;
		unsigned char bytecount = get_char(ptr); ptr += 2;
		unsigned char checksum = bytecount;

		while (address_size)
		{
			address <<= 8;
			address |= get_char(ptr);
			checksum += address & 0xff;
			ptr+=2;
			address_size--;
			bytecount--;
		}

		bytecount--;

		while (bytecount)
		{
			dest[address ^ 1] = get_char(ptr);
			ptr += 2;
			checksum += dest[address ^ 1];
			address++;
			bytecount--;
		}

		unsigned char chk = get_char(ptr);
		checksum ^= 0xff;
		if (chk != checksum)
		{
		//	bprintf (0, _T("checksum failed! Expected: %2.2x Calculated: %2.2x\n"), chk, checksum ^ 0xff);			
			break;
		}
	}

	fclose (fa);
	
	return 0;
}

// remove encryption on 68k rom
static void pgm_decrypt_puzlstar(unsigned char *rom)
{
	int i;
	static const unsigned char puzlstar_tab[256] = {
		0x62, 0x59, 0x17, 0xe3, 0xe1, 0x11, 0x02, 0x97, 0x67, 0x4d, 0x4a, 0x1c, 0x1f, 0x76, 0x64, 0xc1,
		0xfa, 0x04, 0xd2, 0x9f, 0x22, 0xf5, 0xfd, 0xfe, 0x89, 0xeb, 0xab, 0x41, 0xc0, 0xac, 0x2b, 0x64,
		0xfe, 0x1e, 0x9b, 0x68, 0x07, 0xfd, 0x75, 0x25, 0x24, 0xa0, 0x41, 0x5d, 0x79, 0xb5, 0x67, 0x93,
		0xe1, 0xa3, 0x3a, 0x24, 0xa0, 0xbf, 0x64, 0x62, 0x38, 0x65, 0x75, 0x78, 0x8a, 0xf9, 0x7c, 0x00,
		0x71, 0x38, 0xc5, 0xe4, 0xdc, 0xf8, 0xc5, 0xc5, 0x8d, 0x1d, 0x6b, 0xb8, 0x84, 0x4e, 0x21, 0x6d,
		0x55, 0x95, 0x31, 0x86, 0x11, 0xe6, 0xb1, 0x77, 0x77, 0x3b, 0x01, 0x82, 0xab, 0x88, 0xef, 0x77,
		0x08, 0xd8, 0x80, 0xba, 0x41, 0xfb, 0xb3, 0x30, 0x5f, 0xf2, 0x5b, 0xe6, 0x5a, 0x63, 0xb0, 0xa4,
		0x37, 0x37, 0x6d, 0x1f, 0x27, 0x0e, 0x80, 0x8e, 0x1a, 0x45, 0x97, 0xdb, 0x61, 0x10, 0x24, 0x8a,
		0x62, 0x44, 0xf2, 0x91, 0x74, 0x25, 0x3e, 0xc5, 0xe6, 0x8e, 0x55, 0x44, 0x5a, 0x8a, 0x90, 0x25,
		0xa1, 0x0d, 0x43, 0x35, 0x46, 0x80, 0xfe, 0xb3, 0x4e, 0x4c, 0x21, 0xcc, 0x99, 0xc2, 0x3d, 0xce,
		0x19, 0xc5, 0x2c, 0xae, 0xe8, 0x94, 0xad, 0xed, 0x36, 0xb5, 0x8e, 0xf1, 0x81, 0xfe, 0x62, 0x3a,
		0x8c, 0xe0, 0x90, 0x88, 0xf1, 0x31, 0xc5, 0xd9, 0x3c, 0xf4, 0x70, 0x7d, 0x7d, 0x1c, 0xca, 0xe2,
		0x51, 0xce, 0xa0, 0x33, 0x42, 0x1e, 0xf7, 0xe4, 0x13, 0xcf, 0xa0, 0x95, 0x33, 0x59, 0x5b, 0xec,
		0xf7, 0x6d, 0x56, 0xac, 0x6a, 0xae, 0x55, 0x9b, 0x92, 0x14, 0x6d, 0x01, 0x23, 0x4b, 0x80, 0xae,
		0x65, 0x15, 0xc9, 0x84, 0x66, 0xd7, 0xbd, 0x16, 0xd2, 0xe7, 0x15, 0x95, 0x16, 0x9c, 0xd3, 0xd2,
		0x62, 0x51, 0xf6, 0xe3, 0xa6, 0xea, 0x32, 0x86, 0x67, 0xcc, 0x4a, 0x46, 0x4d, 0x37, 0x45, 0xd5
	};

	unsigned short *src = (unsigned short *)rom;

	for (i = 0; i < 0x100000/2; i++) {
		unsigned short x = src[i];

		if ((i & 0x40480) != 0x00080) x ^= 0x0001;
		if ((i & 0x00030) == 0x00010) x ^= 0x0004;
		if ((i & 0x00242) != 0x00042) x ^= 0x0008;
		if ((i & 0x08100) == 0x08000) x ^= 0x0010;
		if ((i & 0x22004) != 0x00004) x ^= 0x0020;
		if ((i & 0x11800) != 0x10000) x ^= 0x0040;
		if ((i & 0x04820) == 0x04820) x ^= 0x0080;

		x ^= puzlstar_tab[i & 0xff] << 8;

		src[i] = x;
	}
}

static void apply_patches(unsigned char *src)
{
	int i, j;
	unsigned short *rom = (unsigned short*)src;

	memcpy (src + 0x100000, src, 0x100000);

	int counter = 0;

	for (i = 0; i < 0x40000/2; i++)
	{	
		if ((rom[i] & 0xff00) == 0x6500 || (rom[i] & 0xff00) == 0x6200) // found bcs / bhi
		{
			unsigned int flags = 0;
			unsigned int offset = 0;
			unsigned int offset_bra = 0;
			unsigned int offset_bcs = 0x100000 + (i * 2);
			unsigned int offset_beq = 0;

			for (j = 0; j < 0x20; j++)
			{
				if ((rom[i-1-j] & 0xff00) == 0x6000 && (rom[i-1-j-2] & 0xfebf) != 0x2c3C) // found bra // 2C3C 
				{
					flags |= 1;
					offset_bra = 0x100000 + ((i-1-j)*2) + 2;
				}
			}
			
			if (flags == 0) continue;
			
			for (j = 0; j < 0x20; j++)
			{
				if ((rom[i+1+j] & 0xff00) == 0x6600) // found bne
				{
					flags |= 2;
				}

				if ((rom[i+1+j] & 0xff00) == 0x6700) // found beq
				{
					if (flags & 2)
					{
						flags |= 4;
						offset = 0x100000 + ((i + j) * 2);
						offset_beq = i + j;
					}
					else
					{
						flags = 0;
					}
					break;
				}
			}

			if (flags == 7){
			
				int alt_cmp = ((rom[i-1] & 0xff00) != 0xff00) ? 1 : 0;
			
				int current_offset = 0xf0000 + (counter * 0x40);

				offset_bra -= 0x100000;
				offset_bcs -= 0x100000;
			
				memset (src + current_offset, 0, 0x40);
			
				int Ax = -1;
				for (j = 0; j < 8; j++)
				{			
					// 3x1y                     move.w  (Ay), Dx
					if ((rom[(offset_bra/2)+j] & 0xf1f8) == 0x3010)
					{
						Ax = rom[(offset_bra/2)+j] & 7;
						break;
					}
					
					// D052                     add.w   (A2), D0
					// 9052                     sub.w   (A2), D0
					if ((rom[(offset_bra/2)+j] & 0xb1f8) == 0x9050)
					{
						Ax = rom[(offset_bra/2)+j] & 7;
						break;
					}
				}

				if (Ax == -1) { /*bprintf (0, _T("error! %x\n"), i * 2 + 0x100000); */continue; }
			
				rom[(current_offset / 2) + 0] = 0xd1fc | (Ax << 9);
				rom[(current_offset / 2) + 1] = 0x10;
				rom[(current_offset / 2) + 2] = 0;
			
				current_offset += 6;
				int copy_len = (offset_bcs + 2) - offset_bra;
			
				memcpy (src + current_offset, src + offset_bra, copy_len);
				current_offset += copy_len;

				if (alt_cmp)
				{
					if (rom[(offset_bcs + 2)/2] == 0x2D6E) { // 2D6E FFF6 FFE8           move.l  (-$a,A6), (-$18,A6)
						memcpy (src + current_offset, src + (offset_bcs + 2), 6);
						current_offset += 6;
						alt_cmp++;
					//	bprintf (0, _T("Extra-%x, %d\n"), offset_bra + copy_len, alt_cmp);
					} else {
						memcpy (src + current_offset, src + offset_bra + copy_len, 2);
						current_offset += 2;
					}
					
					offset_bcs += 2;
				}
			
				rom[(current_offset / 2) + 0] = 0x91fc | (Ax << 9);
				rom[(current_offset / 2) + 1] = 0x10;
				rom[(current_offset / 2) + 2] = 0;
				rom[(current_offset / 2) + 3] = 0x4e75;

				rom[(offset_bcs / 2) - 2] = 0x4eb9;
				current_offset = 0x1f0000 + (counter * 0x40);
				rom[(offset_bcs / 2) - 1] = current_offset >> 16;
				rom[(offset_bcs / 2) - 0] = current_offset & 0xffff;
				if (alt_cmp == 2) rom[(offset_bcs / 2) + 1] = 0x4e71;
				if (alt_cmp == 2) rom[(offset_bcs / 2) + 2] = 0x4e71;
			
			//	bprintf (0, _T("found: (%x) BRA: %5.5x, BCS: %5.5x, count: %d\n"), flags, offset_bra, offset_bcs, counter);
				counter++;
				i = offset_beq;
			}
		}
	}

	for (i = 0; i < 0x100000/2; i++)
	{	
		if (rom[i] == 0x33c0 && rom[i+1] == 0x50 && (rom[i+2] & 0xfffd) == 0)
		{
			int which = rom[i+2] & 2; // 500000 or 500002

			if (which == 0) {
				rom[i] = 0x4e71;
				rom[i+1] = 0x4e71;
				rom[i+2] = 0x4e71;
			} else {
				rom[i] = 0x4eb9;
				rom[i+1] = 0x1f8000 >> 16;
				rom[i+2] = 0x1f8000 & 0xffff;
			}
		}

		//12666C: 33F9 0081 3566 0050 0000 move.w  $813566.l, $500000.l
		//126676: 33F9 0081 3564 0050 0002 move.w  $813564.l, $500002.l
		if (rom[i+0] == 0x33f9 && rom[i+1] == 0x081 && rom[i+2] == 0x8ece && rom[i+3] == 0x50 && rom[i+4] == 0 &&
			rom[i+5] == 0x33f9 && rom[i+6] == 0x081 && rom[i+7] == 0x8ecc && rom[i+8] == 0x50 && rom[i+9] == 2)
		{
			rom[i+0] = 0x4eb9;
			rom[i+1] = 0x1f8000 >> 16;
			rom[i+2] = 0x1f8000 & 0xffff;
			rom[i+3] = rom[i+4] = rom[i+5] = rom[i+6] = rom[i+7] = rom[i+8] = rom[i+9] = 0x4e71;
		}
		
		if ((rom[i] & 0xf1ff) == 0x3039 && rom[i+1] == 0x50 && (rom[i+2] & 0xfffd) == 0)
		{
			int which = rom[i+2] & 2; // 500000 or 500002
			
			rom[i+1] = 0x81e040 >> 16;
			rom[i+2] = (0x81e040 & 0xffff) + (~which & 2); // right?
		}

		//134EBE: 4279 0050 0004           clr.w   $500004.l
		if (rom[i] == 0x4279 && rom[i+1] == 0x50 && rom[i+2] == 4)
		{
			rom[i+0] = rom[i+2] = rom[i+1] = 0x4e71;
		}
	}
}

int main()
{
	printf ("Preparing to build file...\n");

	unsigned char *PGM68KROM = (unsigned char*)malloc(0x400000);

	FILE *fa = fopen("puzzlestar_u2_v100mg.u2", "rb");
	if (fa == NULL) {
		printf ("%s not found!\n", "puzzlestar_u2_v100mg.u2");
		free (PGM68KROM);
		return 1;
	}
	
	FILE *fb = fopen("puzzlestar_u1_v100mg.u1", "rb");
	if (fb == NULL) {
		printf ("%s not found!\n", "puzzlestar_u1_v100mg.u1");
		free (PGM68KROM);
		return 1;
	}
	
	{
		int i;
		for (i = 0; i < 0x100000; i+=2)
		{
			fread (PGM68KROM + i + 0, 1, 1, fa);
			fread (PGM68KROM + i + 1, 1, 1, fb);
		}
	}

	fclose (fa);
	fclose (fb);

	printf ("Program data imported...\n");

	pgm_decrypt_puzlstar(PGM68KROM);
	
	printf ("Program data decrypted...\n");
	
	apply_patches(PGM68KROM);

	printf ("Program data patched...\n");
	
	if (get_s_record(PGM68KROM - 0x100000)) {
		free (PGM68KROM);
		return 1;
	}
	
	fa = fopen("puzlstar_u1-u2-deprotected", "wb");
	fwrite (PGM68KROM, 0x200000, 1, fa);
	fwrite (PGM68KROM, 0x200000, 1, fa);
	fclose (fa);

	free(PGM68KROM);
	
	printf ("complete.\n");
	getch();
	
	return 0;
}
