*-----------------------------------------------------------
* Title         : Puzzle Star protection simulation
* Written by    : IQ_132
* Date          : 2024
* Description   : Simulate the protection routines required by this game in 68k assembly
*-----------------------------------------------------------
    ORG    $1f7ffe
START:                  ; first instruction of program

region_word			dc.w	$0005 ; world

COMMAND_WORD		equ		$818ecc       		; 2 bytes
DATA_WORD		    equ	    $818ece       		; 2 bytes

RESPONSE_VALUE 		equ		$890000

_unused_ram         equ     $81e000
_slots			    equ	    _unused_ram	        ; 40 bytes
_response		    equ	    _unused_ram+$40     ; 4 bytes
_key                equ     _unused_ram+$44     ; 4 bytes
_slot_select		equ	    _unused_ram+$48     ; 2 bytes
_sim_value		    equ	    _unused_ram+$4a		; 2 bytes
_b1_value			equ		_unused_ram+$4c		; 2 bytes
_timer_select		equ		_unused_ram+$4e		; 2 bytes
_timers				equ		_unused_ram+$50		; 8 bytes

_protection_routine
    movem.l     D0-D2/A0, -(A7)

    move.w      D0, D1
    lsr.w       #$8, D1
    cmp.b       #$ff, D1
    bne.s       _skip_calc_key_ff
    	move.l      #$ffffffff, _key
_skip_calc_key_ff
    move.w      _key, D1

    ; apply key to command
    eor.b       D1, D0
    andi.w      #$ff, D0

    ; apply key to sim value
	move.w		DATA_WORD, D2
	eor.w       D1, D2
	move.w      D2, _sim_value

;	debug
;	move.w		DATA_WORD, $500000
;	move.w		COMMAND_WORD, $500002
;	move.w		D0, $500012
;	move.w		D2, $500010

	move.l		#RESPONSE_VALUE, _response

	cmp.b		#$80, D0
	beq			_80_routine
	cmp.b		#$8c, D0
	beq			_8c_routine
	cmp.b		#$90, D0
	beq			_90_routine
	cmp.b		#$99, D0
	beq			_99_routine
	cmp.b		#$b1, D0
	beq			_b1_routine
	cmp.b		#$9d, D0
	beq			_9d_routine
	cmp.b		#$a0, D0
	beq			_a0_routine
	cmp.b		#$ae, D0
	beq			_ae_routine
	cmp.b		#$b0, D0
	beq			_b0_routine
	cmp.b		#$ba, D0
	beq			_ba_routine
	cmp.b		#$bf, D0
	beq			_bf_routine
	cmp.b		#$c1, D0
	beq			_c1_routine
	cmp.b		#$ce, D0
	beq			_ce_routine
	cmp.b		#$cf, D0
	beq			_cf_routine
	cmp.b		#$d0, D0
	beq			_d0_routine
	cmp.b		#$dc, D0
	beq			_dc_routine
	cmp.b		#$e0, D0
	beq			_e0_routine
	cmp.b       #$e5, D0
	beq         _e5_routine	
	cmp.b		#$e7, D0
	beq		    _e7_routine
	cmp.b		#$f8, D0
	beq		    _f8_routine

	; fall through if not found...
_end_protection_routine
    moveq.l     #$0, D0    ; update key!
    moveq.l     #$0, D1
    move.b      _key, D0
    add.w       #$1, D0
    andi.w      #$ff, D0
    cmpi.b      #$ff, D0
    bne.s       _skip_reset_key
    moveq.l     #$1, D0
_skip_reset_key
    move.b      D0, D1
    lsl.w       #$8, D0
    move.b      D1, D0
    move.w      D0, D1
    swap        D0
    move.w      D1, D0
    move.l      D0, _key
    move.l      _response, D1 ; apply key to response!

;	move.w		_response+$2, $500014
;	move.w		_response, $500018
;	move.w		_key, $50001a

    eor.l       D0, D1
    move.l      D1, _response
    movem.l     (A7)+, D0-D2/A0
	rts


_80_routine
	lea.l		_80_lut, A0
	move.w		_sim_value, D0
	andi.l		#$1ff, D0
	add.w		D0, D0
	move.w		(A0,D0.w), D0
	move.l		D0, _response
	bra			_end_protection_routine


_8c_routine
	lea.l		_8c_lut, A0
	move.w		_sim_value, D0
	andi.l		#$3f, D0
	add.w		D0, D0
	move.w		(A0,D0.w), D0
	move.l		D0, _response
	bra			_end_protection_routine


_90_routine
	lea.l		_90_lut, A0
	move.w		_sim_value, D0
	andi.l		#$f, D0
	add.w		D0, D0
	move.w		(A0,D0.w), D0
	move.l		D0, _response
	bra			_end_protection_routine


_99_routine
    move.l      #$01010101, _key
	move.l		#$880000, _response
	bra         _end_protection_routine


_9d_routine
	lea.l		_9d_lut, A0
	move.w		_sim_value, D0
	andi.l		#$f, D0
	add.w		D0, D0
	move.w		(A0,D0.w), D0
	move.l		D0, _response
	bra			_end_protection_routine


_a0_routine
	lea.l		_a0_lut, A0
	move.w		_sim_value, D0
	andi.l		#$f, D0
	add.w		D0, D0
	move.w		(A0,D0.w), D0
	move.l		D0, _response
	bra			_end_protection_routine


_ae_routine
	lea.l		_ae_lut, A0
	move.w		_sim_value, D0
	andi.l		#$f, D0
	add.w		D0, D0
	move.w		(A0,D0.w), D0
	move.l		D0, _response
	bra			_end_protection_routine


_b0_routine
	lea.l		_b0_lut, A0
	move.w		_sim_value, D0
	andi.l		#$f, D0
	add.w		D0, D0
	move.w		(A0,D0.w), D0
	move.l		D0, _response
	bra			_end_protection_routine


_b1_routine
	move.w		_sim_value, _b1_value
	bra			_end_protection_routine


_ba_routine
	lea.l		_ba_lut, A0
	move.w		_sim_value, D0
	andi.l		#$1f, D0
	add.w		D0, D0
	move.w		(A0,D0.w), D0
	move.l		D0, _response
	bra			_end_protection_routine


_bf_routine
	moveq.l		#$0, D0
	moveq.l		#$1, D1
	move.w		_sim_value, D0
	move.w		_b1_value, D1
	mulu.w		D1, D0	; word multiply seems good enough
	move.l		D0, _response
	bra			_end_protection_routine


_d0_routine
	move.l		#$a01000, D0
	moveq.l		#$0, D1
	move.w		_sim_value, D1
	lsl.w		#$5, D1
	add.w		D1, D0
	move.l		D0, _response
	bra			_end_protection_routine


_dc_routine
	move.l		#$a00800, D0
	moveq.l		#$0, D1
	move.w		_sim_value, D1
	lsl.w		#$6, D1
	add.w		D1, D0
	move.l		D0, _response
	bra			_end_protection_routine


_e0_routine
	move.l		#$a00000, D0
	move.w		_sim_value, D0
	andi.w		#$1f, D0
	lsl.w		#$6, D0
	move.l		D0, _response
	bra	        _end_protection_routine


_e5_routine
	move.w		_slot_select, D0
	lea.l		_slots, A0
	move.l		(A0,D0.w), D1
	move.w		_sim_value, D1
	move.l		D1, (A0,D0.w)
	bra         _end_protection_routine


_e7_routine
	move.w		_sim_value, D0
	andi.w      #$f000, D0
	rol.w		#$6, D0
	move.w		D0, _slot_select
	lea.l		_slots, A0
	move.l		(A0,D0.w), D1
	swap		D1
	move.w		_sim_value, D1
	andi.w		#$ff, D1
	swap		D1
	move.l		D1, (A0,D0.w)
	bra         _end_protection_routine


_f8_routine
	move.w		_sim_value, D0
	andi.w		#$f, D0
	lsl.w		#$2, D0
	lea.l		_slots, A0
	move.l		(A0,D0.w), _response
	bra         _end_protection_routine


; timers
_c1_routine
	clr.l		_response
	bra			_end_protection_routine

_ce_routine
	move.w		_sim_value, D0
	add.w		D0, D0
	andi.w		#$6, D0
	move.w		D0, _timer_select
	bra			_end_protection_routine

_cf_routine
	lea.l		_timers, A0
	move.w		_sim_value, D0
	move.w		_timer_select, D1
	move.w		D0, (A0,D1.w)
	bra			_end_protection_routine


_134fe2_routine ; the asic has "timers" just past the shared ram ($4f0000-$4f001f)
	move.l  ($4,A7), D0
	add.l   D0, D0
	addi.l  #_timers, D0 ; was #$4f0020
	movea.l D0, A0
	move.w  (A0), D0
	sub.w	#$1, (A0)
	rts


_initialize	; call from $10023a on init
	movem.l	D0-D1/A0, -(A7)
	lea.l	_unused_ram, A0
	move.w	#$1f, D0
	move.l	#$0, D1
_init_ram_loop
		move.l	D1, (A0)+
	dbra	D0, _init_ram_loop
	movem.l	(A7)+, D0-D1/A0
	jmp		$10091a


_ba_lut	; 1e
	dc.w	$02,$00,$00,$01,$00,$03,$00,$00,$02,$00,$06,$00,$22,$04,$00,$03
	dc.w	$00,$00,$06,$00,$20,$07,$00,$03,$00,$21,$01,$00,$00,$63,$00,$00

_b0_lut ; 10
	dc.w	$09,$0A,$0B,$00,$01,$02,$03,$04,$05,$06,$07,$08,$00,$00,$00,$00

_ae_lut ; 10
	dc.w	$5D,$86,$8C,$8B,$E0,$8B,$62,$AF,$B6,$AF,$10A,$AF,$00,$00,$00,$00

_a0_lut ; 10
	dc.w	$02,$03,$04,$05,$06,$01,$0A,$0B,$0C,$0D,$0E,$09,$00,$00,$00,$00

_9d_lut ; 10
	dc.w	$05,$03,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00

_90_lut ; 10
	dc.w	$0C,$10,$0E,$0C,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00

_8c_lut ; 40
	dc.w	$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$01,$01,$01
	dc.w	$01,$01,$01,$01,$02,$02,$02,$02,$02,$02,$03,$03,$03,$04,$04,$04
	dc.w	$03,$03,$03,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00
	dc.w	$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00

_80_lut ; 1a3
	dc.w	$03,$03,$04,$04,$04,$04,$05,$05,$05,$05,$06,$06,$03,$03,$04,$04
	dc.w	$05,$05,$05,$05,$06,$06,$07,$07,$03,$03,$04,$04,$05,$05,$05,$05
	dc.w	$06,$06,$07,$07,$06,$06,$06,$06,$06,$06,$06,$07,$07,$07,$07,$07
	dc.w	$06,$06,$06,$06,$06,$06,$07,$07,$07,$07,$08,$08,$05,$05,$05,$05
	dc.w	$05,$05,$05,$06,$06,$06,$07,$07,$06,$06,$06,$07,$07,$07,$08,$08
	dc.w	$09,$09,$09,$09,$07,$07,$07,$07,$07,$08,$08,$08,$08,$09,$09,$09
	dc.w	$06,$06,$07,$07,$07,$08,$08,$08,$08,$08,$09,$09,$05,$05,$06,$06
	dc.w	$06,$07,$07,$08,$08,$08,$08,$09,$07,$07,$07,$07,$07,$08,$08,$08
	dc.w	$08,$09,$09,$09,$06,$06,$07,$03,$07,$06,$07,$07,$08,$07,$05,$04
	dc.w	$03,$03,$04,$04,$05,$05,$06,$06,$06,$06,$06,$06,$03,$04,$04,$04
	dc.w	$04,$05,$05,$06,$06,$06,$06,$07,$04,$04,$05,$05,$06,$06,$06,$06
	dc.w	$06,$07,$07,$08,$05,$05,$06,$07,$07,$08,$08,$08,$08,$08,$08,$08
	dc.w	$05,$05,$05,$07,$07,$07,$07,$07,$07,$08,$08,$08,$08,$08,$09,$09
	dc.w	$09,$09,$03,$04,$04,$05,$05,$05,$06,$06,$07,$07,$07,$07,$08,$08
	dc.w	$08,$09,$09,$09,$03,$04,$05,$05,$04,$03,$04,$04,$04,$05,$05,$04
	dc.w	$03,$03,$03,$03,$03,$03,$03,$03,$03,$03,$03,$04,$04,$04,$04,$04
	dc.w	$04,$04,$04,$04,$04,$03,$03,$03,$03,$03,$03,$00,$00,$00,$00,$00
	dc.w	$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00
	dc.w	$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00
	dc.w	$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00
	dc.w	$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00
	dc.w	$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00
	dc.w	$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00
	dc.w	$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00
	dc.w	$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$FF,$00,$00,$00
	dc.w	$00,$FF,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00
	dc.w	$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00


	; 10E508: 13F9 004F 0009 0080 FC3C move.b  $4f0009.l, $80fc3c.l
	ORG $10e50a
	dc.l	region_word+$1

	; 10E5DA: 3039 004F 0008           move.w  $4f0008.l, D0
	ORG $10e5dc
	dc.l	region_word

	ORG $134fe2
	jmp		_134fe2_routine

	ORG $10023a
	dc.l	_initialize

    END    START        ; last line of source





*~Font name~Courier New~
*~Font size~18~
*~Tab type~1~
*~Tab size~4~
